# Copyright 2011,2013,2016 Christoph Reiter
#           2020 Nick Boultbee
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import sys
import time
import os
import traceback
import re
import logging
import errno

from senf import print_, path2fsn, fsn2text, fsnative, supports_ansi_escape_codes

from quodlibet import const
from . import logging as ql_logging

START_TIME = time.time()


class Color:
    NO_COLOR = "\033[0m"
    MAGENTA = "\033[95m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    WHITE = "\033[97m"
    YELLOW = "\033[93m"
    GREEN = "\033[92m"
    RED = "\033[91m"
    BLACK = "\033[90m"
    GRAY = "\033[2m"


class Colorise:
    @classmethod
    def __reset(cls, text):
        return text + Color.NO_COLOR

    @classmethod
    def magenta(cls, text):
        return cls.__reset(Color.MAGENTA + text)

    @classmethod
    def blue(cls, text):
        return cls.__reset(Color.BLUE + text)

    @classmethod
    def cyan(cls, text):
        return cls.__reset(Color.CYAN + text)

    @classmethod
    def white(cls, text):
        return cls.__reset(Color.WHITE + text)

    @classmethod
    def yellow(cls, text):
        return cls.__reset(Color.YELLOW + text)

    @classmethod
    def green(cls, text):
        return cls.__reset(Color.GREEN + text)

    @classmethod
    def red(cls, text):
        return cls.__reset(Color.RED + text)

    @classmethod
    def black(cls, text):
        return cls.__reset(Color.BLACK + text)

    @classmethod
    def bold(cls, text):
        return cls.__reset("\033[1m" + text)

    @classmethod
    def gray(cls, text):
        return cls.__reset(Color.GRAY + text)


_ANSI_ESC_RE = re.compile("(\x1b\\[\\d\\d?m)")
_ANSI_ESC_RE_B = re.compile(b"(\x1b\\[\\d\\d?m)")


def strip_color(text):
    """Strip ansi escape codes from the passed text"""

    if isinstance(text, bytes):
        return _ANSI_ESC_RE_B.sub(b"", text)
    return _ANSI_ESC_RE.sub("", text)


def frame_info(level=0):
    """Return a short string describing the current stack frame which can
    be used for debug messages.

    level defines which frame should be used. 0 means the caller, 1 the caller
    of the caller etc.
    """

    info = ""

    # The frame of the calling function
    if hasattr(sys, "_getframe"):
        frame = sys._getframe()
    else:
        return ""

    for _i in range(level + 1):
        try:
            frame = frame.f_back
        except AttributeError:
            break

    f_code = frame.f_code

    co_name = f_code.co_name
    co_varnames = f_code.co_varnames

    # the calling function got arguments
    if co_varnames and co_varnames[0] in frame.f_locals:
        # the first one could be the class
        cls = frame.f_locals[co_varnames[0]]

        # If it's an instance get the class
        if not hasattr(cls, "__name__"):
            cls = cls.__class__

        # the arg has an attr that is named like the function
        if hasattr(cls, co_name):
            info = cls.__name__

    # else, get the module name
    if not info:
        info = frame.f_globals.get("__name__", "")

    info = str(info)

    # append the function/method name
    if info:
        info += "." + co_name

    return info


def _should_write_to_file(file_):
    """In Windows UI mode we don't have a working stdout/stderr.
    With Python 2 sys.stdout.fileno() returns a negative fd, with Python 3
    sys.stdout is None.
    """

    if file_ is None:
        return False

    try:
        return file_.fileno() >= 0
    except (OSError, AttributeError):
        return True


def _supports_ansi_escape_codes(file_):
    assert file_ is not None
    try:
        return supports_ansi_escape_codes(file_.fileno())
    except (OSError, AttributeError):
        return False


def _print_message(
    string,
    custom_context,
    debug_only,
    prefix,
    color,
    logging_category,
    start_time=START_TIME,
):
    if not isinstance(string, str | fsnative):
        string = str(string)

    context = frame_info(2)

    # strip the package name
    if context.count(".") > 1:
        context = context.split(".", 1)[-1]

    if custom_context:
        context = f"{context}({custom_context!r})"

    timestr = ("%08.3f" % (time.time() - start_time))[-9:]

    color_prefix = getattr(Colorise, color)(prefix)
    info = f"{color_prefix}: [{Colorise.magenta(timestr)}] {Colorise.blue(context)}:"

    lines = string.splitlines()
    if len(lines) > 1:
        string = os.linesep.join([info] + [" " * 4 + l for l in lines])
    else:
        string = info + " " + lines[0]

    if not debug_only or const.DEBUG:
        file_ = sys.stderr
        if _should_write_to_file(file_):
            if not _supports_ansi_escape_codes(file_):
                string = strip_color(string)
            try:
                print_(string, file=file_, flush=True)
            except OSError as e:
                if e.errno == errno.EIO:
                    # When we are started in a terminal with --debug and the
                    # terminal gets closed we lose stdio/err before we stop
                    # printing debug message, resulting in EIO and aborting the
                    # exit process -> Just ignore it.
                    pass
                else:
                    raise

    ql_logging.log(strip_color(string), logging_category)


def format_exception(etype, value, tb, limit=None):
    """Returns a list of str"""

    result_lines = traceback.format_exception(etype, value, tb, limit)
    return [fsn2text(path2fsn(l)) for l in result_lines]


def format_exception_only(etype, value):
    """Returns a list of str"""

    result_lines = traceback.format_exception_only(etype, value)
    return [fsn2text(path2fsn(l)) for l in result_lines]


def format_exc(limit=None):
    """Returns str"""

    etype, value, tb = sys.exc_info()
    return "".join(format_exception(etype, value, tb, limit))


def extract_tb(*args, **kwargs):
    """Returns a list of tuples containing

    (fsnative, int, str, str)
    """

    return traceback.extract_tb(*args, **kwargs)


def print_exc(exc_info=None, context=None):
    """Prints the stack trace of the current exception or the passed one.
    Depending on the configuration will either print a short summary or the
    whole stacktrace.
    """

    if exc_info is None:
        exc_info = sys.exc_info()

    etype, value, tb = exc_info

    if const.DEBUG:
        string = "".join(format_exception(etype, value, tb))
    else:
        # try to get a short error message pointing at the cause of
        # the exception
        text = "".join(format_exception_only(etype, value))
        try:
            filename, lineno, name, line = extract_tb(tb)[-1]
        except IndexError:
            # no stack
            string = text
        else:
            path = fsn2text(path2fsn(os.path.basename(filename)))
            string = f"{path}:{lineno}:{name}: {text}"

    _print_message(string, context, False, "E", "red", "errors")


def print_d(string, context=None):
    """Print debugging information."""

    _print_message(string, context, True, "D", "green", "debug")


def print_w(string, context=None):
    """Print warnings"""

    _print_message(string, context, True, "W", "yellow", "warnings")


def print_e(string, context=None):
    """Print errors"""

    _print_message(string, context, False, "E", "red", "errors")


class PrintHandler(logging.Handler):
    """Converts logging records to our logging format"""

    def emit(self, record):
        print_func = {
            "DEBUG": print_d,
            "INFO": print_d,
            "WARNING": print_w,
            "ERROR": print_e,
            "CRITICAL": print_e,
        }.get(record.levelname, print_d)

        exc_info = record.exc_info
        context = f"{record.module}.{record.funcName}"
        record.exc_info = None
        print_func(self.format(record), context=context)
        if exc_info is not None:
            print_exc(record.exc_info, context=context)



# Copyright 2011-2021 Nick Boultbee
#
# Inspired in parts by PySqueezeCenter (c) 2010 JingleManSweep
# SqueezeCenter and SqueezeBox are copyright Logitech
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

from quodlibet import _
from quodlibet import qltk, app
from quodlibet.ext._shared.squeezebox.base import SqueezeboxPluginMixin
from quodlibet.util.dprint import print_d
from quodlibet.qltk import Icons
from quodlibet.plugins.events import EventPlugin

if os.name == "nt":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError


class SqueezeboxSyncPlugin(EventPlugin, SqueezeboxPluginMixin):
    PLUGIN_ID = "Squeezebox Output"
    PLUGIN_NAME = _("Squeezebox Sync")
    PLUGIN_DESC_MARKUP = (
        _(
            "Makes Logitech Squeezebox mirror Quod Libet output, "
            "provided both read from an identical library."
        )
        + "\n"
        + _(
            'Shares configuration with <a href="%(plugin_link)s">Export to '
            "Squeezebox plugin</a>."
        )
        % {"plugin_link": "quodlibet:///prefs/plugins/Export to Squeezebox Playlist"}
    )
    PLUGIN_ICON = Icons.MEDIA_PLAYBACK_START

    server = None
    active = False
    _debug = False

    def __init__(self):
        super().__init__()

    @classmethod
    def post_reconnect(cls):
        cls.server.stop()
        SqueezeboxPluginMixin.post_reconnect()
        player = app.player
        cls.plugin_on_song_started(player.info)
        cls.plugin_on_seek(player.info, player.get_position())

    def enabled(self):
        print_d(f"Debug is set to {self._debug}")
        self.active = True
        self.init_server()
        self.server.pause()
        if not self.server.is_connected:
            qltk.ErrorMessage(
                None,
                _("Error finding Squeezebox server"),
                _("Error finding %s. Please check settings") % self.server.config,
            ).run()

    def disabled(self):
        # Stopping might be annoying in some situations, but seems more correct
        if self.server:
            self.server.stop()
        self.active = False

    @classmethod
    def plugin_on_song_started(cls, song):
        # Yucky hack to allow some form of immediacy on re-configuration
        cls.server._debug = cls._debug = cls.config_get_bool("debug", False)
        if cls._debug:
            print_d("Paused" if app.player.paused else "Not paused")
        if song and cls.server and cls.server.is_connected:
            path = cls.get_sb_path(song)
            print_d(f"Requesting to play {path}...")
            if app.player.paused:
                cls.server.change_song(path)
            else:
                cls.server.playlist_play(path)

    @classmethod
    def plugin_on_paused(cls):
        if cls.server:
            cls.server.pause()

    @classmethod
    def plugin_on_unpaused(cls):
        if cls.server:
            cls.server.unpause()

    @classmethod
    def plugin_on_seek(cls, song, msec):
        if not app.player.paused:
            if cls.server:
                cls.server.seek_to(msec)
                cls.server.play()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import time
from pathlib import Path
from collections.abc import Generator, Iterable

from gi.repository import Gio, GLib, GObject

from quodlibet import print_d, print_w, _, formats
from quodlibet.formats import AudioFileError, AudioFile
from quodlibet.library.base import iter_paths, Library, PicklingMixin
from quodlibet.qltk.notif import Task
from quodlibet.util import copool, print_exc
from quodlibet.util.library import get_exclude_dirs
from quodlibet.util.path import ismount, unexpand, normalize_path
from senf import fsn2text, fsnative


class FileLibrary(Library[fsnative, AudioFile], PicklingMixin):
    """A library containing items on a local(-ish) filesystem.

    These must support the valid, exists, mounted, and reload methods,
    and have a mountpoint attribute.
    """

    def __init__(self, name=None):
        super().__init__(name)
        self._masked = {}

    def _load_init(self, items):
        """Add many items to the library, check if the
        mountpoints are available and mark items as masked if not.

        Does not check if items are valid.
        """

        mounts = {}
        contents = self._contents
        masked = self._masked

        for item in items:
            mountpoint = item.mountpoint

            if mountpoint not in mounts:
                is_mounted = ismount(mountpoint)

                # In case mountpoint is mounted through autofs we need to
                # access a sub path for it to mount
                # https://github.com/quodlibet/quodlibet/issues/2146
                if not is_mounted:
                    item.exists()
                    is_mounted = ismount(mountpoint)

                mounts[mountpoint] = is_mounted
                # at least one not mounted, make sure masked has an entry
                if not is_mounted:
                    masked.setdefault(mountpoint, {})

            if mounts[mountpoint]:
                contents[item.key] = item
            else:
                masked[mountpoint][item.key] = item

    def _load_item(self, item, force=False):
        """Add an item, or refresh it if it's already in the library.
        No signals will be fired.
        Return a tuple of booleans: (changed, removed)
        """
        print_d(f"Loading {item.key!r}", self._name)
        valid = item.valid()

        # The item is fine; add it if it's not present.
        if not force and valid:
            print_d(f"{item.key!r} is valid.", self._name)
            self._contents[item.key] = item
            return False, False
        # Either we should force a load, or the item is not okay.
        # We're going to reload; this could change the key.  So
        # remove the item if it's currently in.
        try:
            del self._contents[item.key]
        except KeyError:
            present = False
        else:
            present = True
        # If the item still exists, reload it.
        if item.exists():
            try:
                item.reload()
            except AudioFileError:
                print_w(f"Error reloading {item.key!r}", self._name)
                return False, True
            else:
                print_d(f"Reloaded {item.key!r}.", self._name)
                self._contents[item.key] = item
                return True, False
        elif not item.mounted():
            # We don't know if the item is okay or not, since
            # it's not not mounted. If the item was present
            # we need to mark it as removed.
            print_d(f"Masking {item.key!r}", self._name)
            self._masked.setdefault(item.mountpoint, {})
            self._masked[item.mountpoint][item.key] = item
            return False, present
        else:
            # The item doesn't exist at all anymore. Mark it as
            # removed if it was present, otherwise nothing.
            print_d(f"Ignoring (so removing) {item.key!r}.", self._name)
            return False, present

    def reload(self, item, changed=None, removed=None):
        """Reload a song, possibly noting its status.

        If sets are given, it assumes the caller will handle signals,
        and only updates the sets. Otherwise, it handles signals
        itself. It *always* handles library contents, so do not
        try to remove (again) a song that appears in the removed set.
        """

        was_changed, was_removed = self._load_item(item, force=True)
        assert not (was_changed and was_removed)

        if was_changed:
            if changed is None:
                self.emit("changed", {item})
            else:
                changed.add(item)
        elif was_removed:
            if removed is None:
                self.emit("removed", {item})
            else:
                removed.add(item)

    def rebuild(self, paths, force=False, exclude=None, cofuncid=None):
        """Reload or remove songs if they have changed or been deleted.

        This generator rebuilds the library over the course of iteration.

        Any paths given will be scanned for new files, using the 'scan'
        method.

        Only items present in the library when the rebuild is started
        will be checked.

        If this function is copooled, set "cofuncid" to enable pause/stop
        buttons in the UI.
        """

        print_d(f"Rebuilding, force is {force}", self._name)

        task = Task(_("Library"), _("Checking mount points"))
        if cofuncid:
            task.copool(cofuncid)
        for _i, (point, items) in task.list(enumerate(self._masked.items())):
            if ismount(point):
                self._contents.update(items)
                del self._masked[point]
                self.emit("added", list(items.values()))
                yield True

        task = Task(_("Library"), _("Scanning library"))
        if cofuncid:
            task.copool(cofuncid)
        changed, removed = set(), set()
        for i, (key, item) in task.list(enumerate(sorted(self.items()))):
            if key in self._contents and force or not item.valid():
                self.reload(item, changed, removed)
                # These numbers are pretty empirical. We should yield more
            # often than we emit signals; that way the main loop stays
            # interactive and doesn't get bogged down in updates.
            if len(changed) >= 200:
                self.emit("changed", changed)
                changed = set()
            if len(removed) >= 200:
                self.emit("removed", removed)
                removed = set()
            if len(changed) > 20 or i % 200 == 0:
                yield True
        print_d(f"Removing {len(removed)}, changing {len(changed)}).", self._name)
        if removed:
            self.emit("removed", removed)
        if changed:
            self.emit("changed", changed)

        yield from self.scan(paths, exclude, cofuncid)

    def add_filename(self, filename: str | Path, add: bool = True) -> AudioFile | None:
        """Add a file based on its filename.
        Subclasses must override this to open the file correctly.

        :return: the audio file if added (or None)
        """

    def contains_filename(self, filename) -> bool:
        """Returns if a song for the passed filename is in the library."""
        key = normalize_path(filename, True)
        return key in self._contents

    def scan(
        self,
        paths: Iterable[fsnative],
        exclude: Iterable[fsnative] | None = None,
        cofuncid=None,
    ):
        def need_yield(last_yield=[0]):  # noqa
            current = time.time()
            if abs(current - last_yield[0]) > 0.015:
                last_yield[0] = current
                return True
            return False

        def need_added(last_added=[0]):  # noqa
            current = time.time()
            if abs(current - last_added[0]) > 1.0:
                last_added[0] = current
                return True
            return False

        # first scan each path for new files
        paths_to_load = []
        for scan_path in paths:
            print_d(f"Scanning {scan_path}", self._name)
            desc = _("Scanning %s") % (fsn2text(unexpand(scan_path)))
            with Task(_("Library"), desc) as task:
                if cofuncid:
                    task.copool(cofuncid)

                for real_path in iter_paths(scan_path, exclude=exclude):
                    if need_yield():
                        task.pulse()
                        yield
                    # skip unknown file extensions
                    if not formats.filter(real_path):
                        continue
                    # already loaded
                    if self.contains_filename(real_path):
                        continue
                    paths_to_load.append(real_path)

        yield

        # then (try to) load all new files
        with Task(_("Library"), _("Loading files")) as task:
            if cofuncid:
                task.copool(cofuncid)

            added = []
            for real_path in task.gen(paths_to_load):
                item = self.add_filename(real_path, False)
                if item is not None:
                    added.append(item)
                    if len(added) > 100 or need_added():
                        self.add(added)
                        added = []
                        yield
                if added and need_yield():
                    yield
            if added:
                self.add(added)
                added = []
                yield True

    def get_content(self):
        """Return visible and masked items"""

        items = list(self.values())
        for masked in self._masked.values():
            items.extend(masked.values())

        # Item keys are often based on filenames, in which case
        # sorting takes advantage of the filesystem cache when we
        # reload/rescan the files.
        items.sort(key=lambda item: item.key)

        return items

    def masked(self, item):
        """Return true if the item is in the library but masked."""
        try:
            point = item.mountpoint
        except AttributeError:
            # Checking a key.
            for point in self._masked.values():
                if item in point:
                    return True
        else:
            # Checking a full item.
            return item in self._masked.get(point, {}).values()

    def unmask(self, point):
        print_d(f"Unmasking {point!r}", self._name)
        items = self._masked.pop(point, {})
        if items:
            self.add(items.values())

    def mask(self, point):
        print_d(f"Masking {point!r}", self._name)
        removed = {}
        for item in self.values():
            if item.mountpoint == point:
                removed[item.key] = item
        if removed:
            self.remove(removed.values())
            self._masked.setdefault(point, {}).update(removed)

    @property
    def masked_mount_points(self):
        """List of mount points that contain masked items"""

        return list(self._masked.keys())

    def get_masked(self, mount_point):
        """List of items for a mount point"""

        return list(self._masked.get(mount_point, {}).values())

    def remove_masked(self, mount_point):
        """Remove all songs for a masked point"""

        self._masked.pop(mount_point, {})

    def move_root(
        self, old_root: str, new_root: fsnative, write_files: bool = True
    ) -> Generator[None, None, None]:
        """
        Move the root for all songs in a given (scan) directory.

        We avoid dereferencing the destination, to allow users things like:
          1. Symlink new_path -> old_root
          2. Move QL root to new_path
          3. Remove symlink
          4. Move audio files: old_root -> new_path

        """
        # TODO: only move primary library
        old_path = Path(normalize_path(old_root, canonicalise=True)).expanduser()
        new_path = Path(normalize_path(new_root)).expanduser()
        if not old_path.is_dir():
            print_w(
                f"Source dir {str(old_path)!r} doesn't exist, assuming that's OK",
                self._name,
            )
        if not new_path.is_dir():
            raise ValueError(f"Destination {new_path!r} is not a directory")
        print_d(f"Checking entire library for {str(old_path)!r}", self._name)
        missing: set[AudioFile] = set()
        changed = set()
        total = len(self)
        if not total:
            return
        with Task(_("Library"), _("Moving library files")) as task:
            yield
            for i, song in enumerate(list(self.values())):
                task.update(i / total)
                key = normalize_path(song.key)
                path = Path(key)
                if old_path in path.parents:
                    # TODO: more Pathlib-friendly dir replacement...
                    new_key = key.replace(str(old_path), str(new_path), 1)
                    new_key = normalize_path(new_key, canonicalise=False)
                    if new_key == key:
                        print_w(f"Substitution failed for {key!r}", self._name)
                    # We need to update ~filename and ~mountpoint
                    song.sanitize()
                    if write_files:
                        song.write()
                    if self.move_song(song, new_key):
                        changed.add(song)
                    else:
                        missing.add(song)
                elif not (i % 1000):
                    print_d(f"Not moved, for example: {key!r}", self._name)
                if not i % 100:
                    yield
            self.changed(changed)
            if missing:
                print_w(
                    f"Couldn't find {len(list(missing))} files: {missing}", self._name
                )
        yield
        self.save()
        print_d(
            f"Done moving {len(changed)} track(s) (of {total}) "
            f"to {str(new_path)!r}.",
            self._name,
        )

    def move_song(self, song: AudioFile, new_path: fsnative) -> bool:
        """Updates the location of a song, without touching the file.

        :returns: True if it was could be found (and moved)
        """
        existed = True
        key = song.key
        print_d(f"Moving {key!r} -> {new_path!r}", self._name)
        try:
            del self._contents[key]  # type: ignore
        except KeyError:
            existed = False
            # Continue - maybe it's already moved
        song.sanitize(new_path)
        self._contents[new_path] = song
        return existed

    def remove_roots(self, old_roots: Iterable[str]) -> Generator[None, None, None]:
        """Remove library roots (scandirs) entirely, and all their songs"""
        old_paths = [
            Path(normalize_path(root, canonicalise=True)).expanduser()
            for root in old_roots
        ]
        total = len(self)
        removed = set()
        print_d(f"Removing library roots {old_roots}", self._name)
        yield
        with Task(_("Library"), _("Removing library files")) as task:
            for i, song in enumerate(list(self.values())):
                task.update(i / total)
                key = normalize_path(song.key)
                song_path = Path(key)
                if any(path in song_path.parents for path in old_paths):
                    removed.add(song)
                if not i % 100:
                    yield
        if removed:
            self.remove(removed)
        else:
            print_d(f"No tracks in {old_roots} to remove", self._name)


Event = Gio.FileMonitorEvent


class WatchedFileLibraryMixin(FileLibrary):
    """A File Library that sets up monitors on directories at refresh
    and handles changes sensibly"""

    def __init__(self, name=None):
        super().__init__(name)
        self._monitors: dict[Path, tuple[GObject.GObject, int]] = {}

    def monitor_dir(self, path: Path) -> None:
        """Monitors a single directory"""

        # Only add one monitor per absolute path...
        if path not in self._monitors:
            f = Gio.File.new_for_path(str(path))
            try:
                monitor = f.monitor_directory(Gio.FileMonitorFlags.WATCH_MOVES, None)
            except GLib.GError as e:
                print_w(f"Couldn't watch {path} ({e})", self._name)
                monitor = None
            if not monitor:
                return
            handler_id = monitor.connect("changed", self.__file_changed)
            # Don't destroy references - http://stackoverflow.com/q/4535227
            self._monitors[path] = (monitor, handler_id)
            print_d(f"Monitoring {path!s}", self._name)

    def __file_changed(
        self,
        _monitor,
        main_file: Gio.File,
        other_file: Gio.File | None,
        event: Gio.FileMonitorEvent,
    ) -> None:
        if event == Event.CHANGES_DONE_HINT:
            # This seems to work fine on most Linux, but not on Windows / macOS
            # Or at least, not in CI anyway.
            # So shortcut the whole thing
            return
        try:
            file_path = main_file.get_path()
            if file_path is None:
                return
            file_path = normalize_path(file_path, True)
            song = self.get(file_path)
            file_path = Path(file_path)
            other_path = (
                Path(normalize_path(other_file.get_path(), True))
                if other_file
                else None
            )
            if event in (Event.CREATED, Event.MOVED_IN):
                if file_path.is_dir():
                    self.monitor_dir(file_path)
                    copool.add(self.scan, [str(file_path)])
                elif not song:
                    print_d(f"Auto-adding created file: {file_path}", self._name)
                    self.add_filename(str(file_path))
            elif event == Event.RENAMED:
                if not other_path:
                    print_w(
                        f"No destination found for rename of {file_path}", self._name
                    )
                if song:
                    print_d(f"Moving {file_path} to {other_path}...", self._name)
                    if self.move_song(song, str(other_path)):  # type:ignore
                        print_w(f"Song {file_path} has gone")
                elif self.is_monitored_dir(file_path):
                    if self.librarian:
                        print_d(
                            f"Moving tracks from {file_path} -> {other_path}...",
                            self._name,
                        )
                        copool.add(
                            self.librarian.move_root,
                            str(file_path),
                            str(other_path),
                            write_files=False,
                            priority=GLib.PRIORITY_DEFAULT,
                        )
                    self.unmonitor_dir(file_path)
                    if other_path:
                        self.monitor_dir(other_path)
                else:
                    if other_path:
                        print_w(
                            f"Seems {file_path} is not a track (deleted?)", self._name
                        )
                        # On some (Windows?) systems CHANGED is called which can remove
                        # before we get here, so let's try adding the new path back
                        self.add_filename(other_path)

            elif event == Event.CHANGED:
                if song:
                    # QL created (or knew about) this one; still check if it changed
                    if not song.valid():
                        self.reload(song)
                else:
                    print_d(f"Auto-adding new file: {file_path}", self._name)
                    self.add_filename(file_path)
            elif event in (Event.MOVED_OUT, Event.DELETED):
                if song:
                    print_d(f"...so deleting {file_path}", self._name)
                    self.reload(song)
                else:
                    # either not a song, or a song that was renamed by QL
                    if self.is_monitored_dir(file_path):
                        self.unmonitor_dir(file_path)

                    # And try to remove all songs under that dir. Slowly.
                    gone = set()
                    for key, song in self.iteritems():
                        if file_path in Path(key).parents:
                            gone.add(song)
                    if gone:
                        print_d(
                            f"Removing {len(gone)} contained songs in {file_path}",
                            self._name,
                        )
                        actually_gone = self.remove(gone)
                        if gone != actually_gone:
                            print_w(
                                f"Couldn't remove all: {gone - actually_gone}",
                                self._name,
                            )
            else:
                print_d(
                    f"Unhandled event {event} on {file_path} ({other_path})", self._name
                )
                return
        except Exception:
            print_w("Failed to run file monitor callback", self._name)
            print_exc()
        print_d(f"Finished handling {event}", self._name)

    def is_monitored_dir(self, path: Path) -> bool:
        return path in self._monitors

    def unmonitor_dir(self, path: Path) -> None:
        """Disconnect and remove any monitor for a directory, if found"""

        monitor, handler_id = self._monitors.get(path, (None, None))
        if not monitor:
            print_d(f"Couldn't find path {path} in active monitors", self._name)
            return
        monitor.disconnect(handler_id)
        del self._monitors[path]

    def start_watching(self, paths: Iterable[fsnative]):
        print_d(f"Setting up file watches on {paths}...", self._name)
        exclude_dirs = [e for e in get_exclude_dirs() if e]

        def watching_producer():
            # TODO: integrate this better with scanning.
            for fullpath in paths:
                desc = _("Adding watches for %s") % (fsn2text(unexpand(fullpath)))
                with Task(_("Library"), desc) as task:
                    normalised = Path(normalize_path(fullpath, True)).expanduser()
                    if any(
                        Path(exclude) in normalised.parents for exclude in exclude_dirs
                    ):
                        continue
                    unpulsed = 0
                    self.monitor_dir(normalised)
                    for path, dirs, _files in os.walk(normalised):
                        normalised = Path(normalize_path(path, True))
                        for d in dirs:
                            self.monitor_dir(normalised / d)
                        unpulsed += len(dirs)
                        if unpulsed > 50:
                            task.pulse()
                            unpulsed = 0
                        yield

        copool.add(watching_producer, funcid="watch_library")

    def stop_watching(self):
        print_d(f"Removing watches on {len(self._monitors)} dirs", self._name)

        for monitor, handler_id in self._monitors.values():
            monitor.disconnect(handler_id)
        self._monitors.clear()

    def destroy(self):
        self.stop_watching()
        super().destroy()



# Copyright 2010 Christoph Reiter
#        2016,20 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _
from quodlibet import app
from quodlibet.order import OrderInOrder, OrderRemembered
from quodlibet.plugins.playorder import ShufflePlugin
from quodlibet.qltk import Icons


class FollowOrder(ShufflePlugin, OrderInOrder, OrderRemembered):
    PLUGIN_ID = "follow"
    PLUGIN_NAME = _("Follow Cursor")
    PLUGIN_ICON = Icons.GO_JUMP
    PLUGIN_DESC = _(
        "Adds a play order mode that follows your selection, "
        "or the next song in the list once exhausted."
    )
    display_name = _("Follow cursor")
    accelerated_name = _("_Follow cursor")

    __last_path = None

    def next(self, playlist, iter):
        next_fallback = OrderInOrder.next(self, playlist, iter)
        OrderRemembered.next(self, playlist, iter)

        selected = app.window.songlist.get_selected_songs()
        if not selected:
            return next_fallback

        selected_iter = playlist.find(selected[0])
        selected_path = playlist.get_path(selected_iter)
        current_path = iter and playlist.get_path(iter)

        if selected_path in (current_path, self.__last_path):
            return next_fallback

        self.__last_path = selected_path
        return selected_iter

    def previous(self, *args):
        self.__last_path = None
        return super().previous(*args)

    def set(self, playlist, iter):
        if iter:
            self.__last_path = playlist.get_path(iter)
        return super().set(playlist, iter)

    def reset(self, playlist):
        super().reset(playlist)
        self.__last_path = None



# Copyright 2012-2016 Ryan "ZDBioHazard" Turner <zdbiohazard2@gmail.com>
#           2016-2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import math
import random

from gi.repository import Gtk

from quodlibet import _, print_d
from quodlibet.order.reorder import Reorder
from quodlibet.plugins import PluginConfigMixin
from quodlibet.plugins.playorder import ShufflePlugin
from quodlibet.order import OrderRemembered
from quodlibet.qltk import Icons


class PlaycountEqualizer(ShufflePlugin, OrderRemembered, PluginConfigMixin):
    PLUGIN_ID = "playcounteq"
    PLUGIN_NAME = _("Playcount Equalizer")
    PLUGIN_DESC = _("Adds a shuffle mode that prefers songs with fewer total plays.")
    PLUGIN_ICON = Icons.MEDIA_PLAYLIST_SHUFFLE
    display_name = _("Prefer less played")
    accelerated_name = _("Prefer _less played")

    priority = Reorder.priority

    _MAGNITUDE_DEFAULT = 1

    # Select the next track.
    def next(self, playlist, current):
        super().next(playlist, current)

        remaining = self.remaining(playlist)

        # Don't try to search through an empty / played playlist.
        if len(remaining) <= 0:
            return None

        mag_cfg = float(self.config_get("magnitude", self._MAGNITUDE_DEFAULT))
        # Adjusting input to range from 1 to 3
        # weights will be calculated as a power of this value.
        magn = (mag_cfg * 2.0) / 100.0 + 1.0

        # Set-up the search information.
        max_count = max([song("~#playcount") for song in remaining.values()])

        weights = {
            i: math.ceil(math.pow(max_count - song("~#playcount") + 1, magn))
            for i, song in remaining.items()
        }

        weights_sum = sum(weights.values())
        choice = int(max(1, math.ceil(weights_sum * random.random())))

        print_d(f"Weighted random: {choice} for the weights sum: {weights_sum}")

        # Search for a track.
        for i, weight in weights.items():
            choice -= weight
            if choice <= 0:
                return playlist.get_iter([i])
        else:  # This should only happen if all songs have equal play counts.
            return playlist.get_iter([random.choice(list(remaining.keys()))])

    @classmethod
    def PluginPreferences(cls, parent):
        def magnitude_changed(spin):
            cls.config_set("magnitude", int(spin.get_value_as_int()))

        vb = Gtk.VBox(spacing=10)
        vb.set_border_width(10)
        hbox = Gtk.HBox(spacing=6)
        lbl = Gtk.Label(label=_("Priority for less played tracks"))
        hbox.pack_start(lbl, False, True, 0)

        val = cls.config_get("magnitude", cls._MAGNITUDE_DEFAULT)

        spin = Gtk.SpinButton(adjustment=Gtk.Adjustment.new(int(val), 1, 100, 1, 10, 0))
        spin.connect("value-changed", magnitude_changed)
        hbox.pack_start(spin, False, True, 0)
        vb.pack_start(hbox, True, True, 0)
        vb.show_all()
        return vb



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from datetime import datetime
from time import localtime, strftime


def format_date(seconds: float, format_setting: str | None = None) -> str:
    """Formats a date either with the default format,
    or the passed strftime-compatible format string"""
    try:
        date = datetime.fromtimestamp(seconds).date()
    except (OverflowError, ValueError, OSError):
        text = ""
    else:
        if format_setting:
            format_ = format_setting
        # use default behaviour-format
        else:
            today = datetime.now().date()
            days = (today - date).days
            if days == 0:
                format_ = "%X"
            elif days < 7:
                format_ = "%A"
            else:
                format_ = "%x"

        stamp = localtime(seconds)
        text = strftime(format_, stamp)
    return text



# Copyright 2005-2010   Joshua Kwan <joshk@triplehelix.org>,
#                       Michael Ball <michael.ball@gmail.com>,
#                       Steven Robertson <steven@strobe.cc>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet.plugins import PluginConfig


def get_config():
    pc = PluginConfig("brainz")

    defaults = pc.defaults
    defaults.set("year_only", False)
    defaults.set("albumartist", True)
    defaults.set("artist_sort", False)
    defaults.set("standard", True)
    defaults.set("labelid2", False)

    return pc


pconfig = get_config()



# Copyright 2009-2014 Christoph Reiter
#             2021-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import hashlib
from pathlib import Path
from tempfile import gettempdir

from gi.repository import GdkPixbuf, GLib

from quodlibet import print_w, print_d
from senf import fsn2uri, fsnative

import quodlibet
from quodlibet.util.path import mtime, mkdir, xdg_get_cache_home
from quodlibet.util import enum
from quodlibet.qltk.image import scale


def get_thumbnail_folder() -> Path:
    """Returns a path to the thumbnail folder.

    The returned path might not exist.
    """

    if os.name == "nt":
        thumb_folder = Path(quodlibet.get_cache_dir()) / "thumbnails"
    else:
        cache_folder = Path(xdg_get_cache_home()) / "thumbnails"
        thumb_folder = Path("~/.thumbnails").expanduser()
        if cache_folder.exists() or not thumb_folder.exists():
            thumb_folder = cache_folder

    return thumb_folder


@enum
class ThumbSize(int):
    NORMAL = 128
    LARGE = 256
    LARGEST = LARGE


def get_cache_info(path: Path, boundary: tuple[int, int]) -> tuple[Path, int]:
    """For an image at `path` return (cache_path, thumb_size)

    cache_path points to a potential cache file
    thumb size is either 128 or 256
    """

    width, height = boundary

    if width <= ThumbSize.NORMAL and height <= ThumbSize.NORMAL:
        size_name = "normal"
        thumb_size = ThumbSize.NORMAL
    else:
        size_name = "large"
        thumb_size = ThumbSize.LARGE

    thumb_folder = get_thumbnail_folder()
    cache_dir = thumb_folder / size_name

    uri = fsn2uri(str(path))
    thumb_name = hashlib.md5(uri.encode("ascii")).hexdigest() + ".png"
    thumb_path = cache_dir / thumb_name

    return thumb_path, thumb_size


def get_thumbnail_from_file(fileobj, boundary) -> GdkPixbuf.Pixbuf | None:
    """Like get_thumbnail() but works with files that can't be reopened.

    This is needed on Windows where NamedTemporaryFile can't be reopened.

    :returns: Pixbuf or None. Thread-safe.
    """

    assert fileobj

    try:
        path = fileobj.name
        assert isinstance(path, fsnative), path
        return get_thumbnail(path, boundary)
    except GLib.GError:
        try:
            loader = GdkPixbuf.PixbufLoader()
            loader.set_size(*boundary)
            loader.write(fileobj.read())
            loader.close()
            fileobj.seek(0, 0)
            # can return None in case of partial data
            return loader.get_pixbuf()
        except (OSError, GLib.GError) as e:
            print_w(f"Couldn't load thumbnail with PixbufLoader either: {e}")
    return None


def get_thumbnail(path: fsnative, boundary, ignore_temp=True) -> GdkPixbuf:
    """Get a thumbnail pixbuf of an image at `path`.

    Will create/use a thumbnail in the user's thumbnail directory if possible.
    Follows the Free Desktop specification:

    http://specifications.freedesktop.org/thumbnail-spec/

    If ignore_temp then no thumbnail cache will be created for files
    in the default temporary directory.

    Can raise GLib.GError. Thread-safe.
    """

    assert isinstance(path, fsnative), type(path)

    width, height = boundary
    new_from_file_at_size = GdkPixbuf.Pixbuf.new_from_file_at_size

    # larger than thumbnails, load directly
    if width > ThumbSize.LARGEST or height > ThumbSize.LARGEST:
        return new_from_file_at_size(path, width, height)

    path_mtime = mtime(path)
    if path_mtime == 0:
        return new_from_file_at_size(path, width, height)

    # embedded thumbnails come from /tmp/
    # FIXME: move this to another layer
    if ignore_temp and path.startswith(gettempdir()):
        return new_from_file_at_size(path, width, height)

    thumb_path, thumb_size = get_cache_info(Path(path), boundary)
    cache_dir = thumb_path.parent
    try:
        mkdir(cache_dir, 0o700)
    except OSError as e:
        print_w(f"Couldn't create cache dir {cache_dir!r} ({e}")
        return new_from_file_at_size(path, width, height)

    try:
        pb = new_from_file_at_size(str(thumb_path), width, height)
    except GLib.GError:
        # in case it fails to load, we recreate it
        print_w(f"Couldn't find thumbnail at {str(thumb_path)!r}, so recreating.")
    else:
        meta_mtime = pb.get_option("tEXt::Thumb::MTime")
        if meta_mtime is not None:
            try:
                meta_mtime = int(meta_mtime)
            except ValueError:
                pass
            else:
                if meta_mtime == int(path_mtime):
                    return pb

    info, pw, ph = GdkPixbuf.Pixbuf.get_file_info(path)

    # Too small picture, no thumbnail needed
    if pw < thumb_size and ph < thumb_size:
        return new_from_file_at_size(path, width, height)

    thumb_pb = new_from_file_at_size(path, thumb_size, thumb_size)

    uri = fsn2uri(path)
    mime = info.get_mime_types()[0]
    options = {
        "tEXt::Thumb::Image::Width": str(pw),
        "tEXt::Thumb::Image::Height": str(ph),
        "tEXt::Thumb::URI": uri,
        "tEXt::Thumb::MTime": str(int(path_mtime)),
        "tEXt::Thumb::Size": str(os.path.getsize(path)),
        "tEXt::Thumb::Mimetype": mime,
        "tEXt::Software": "QuodLibet",
    }

    print_d(f"Saving thumbnail to {thumb_path!s}")
    thumb_pb.savev(str(thumb_path), "png", list(options.keys()), list(options.values()))
    try:
        thumb_path.chmod(0o600)
    except OSError:
        pass

    return scale(thumb_pb, boundary)



# Copyright 2017 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Interface around faulthandler to save and restore segfault info for the
next program invocation.
"""

import os
import ctypes
import errno
import re
import atexit

import faulthandler

from quodlibet.util.dprint import print_exc


_fileobj = None


class FaultHandlerCrash(Exception):  # noqa
    """The exception type used for raising errors with a faulthandler
    stacktrace. Needed so we can add special handling in the error reporting
    code paths.
    """

    def get_grouping_key(self):
        """Given a stacktrace produced by the faulthandler module returns a
        short string for grouping similar stacktraces together.

        Args:
            stacktrace (str)
        Returns:
            str
        """

        stacktrace = str(self)
        if isinstance(stacktrace, bytes):
            stacktrace = stacktrace.decode("utf-8", "replace")

        assert isinstance(stacktrace, str)

        # Extract the basename and the function name for each line and hash
        # them. Could be smarter, but let's try this for now..
        reg = re.compile(r'.*?"([^"]+).*?(\w+$)')
        values = []
        for l in stacktrace.splitlines():
            m = reg.match(l)
            if m is not None:
                path, func = m.groups()
                path = os.path.basename(path)
                values.extend([path, func])
        return "|".join(values)


def enable(path):
    """Enable crash reporting and create empty target file

    Args:
        path (pathlike): the location of the crash log target path
    Raises:
        IOError: In case the location is not writable
    """

    global _fileobj

    if _fileobj is not None:
        raise Exception("already enabled")

    # we open as reading so raise_and_clear_error() can extract the old error
    try:
        _fileobj = open(path, "rb+")
    except OSError as e:
        if e.errno == errno.ENOENT:
            _fileobj = open(path, "wb+")
        else:
            raise

    faulthandler.enable(_fileobj, all_threads=False)


def disable():
    """Disable crash reporting and removes the target file

    Does not raise.
    """

    global _fileobj

    if _fileobj is None:
        return

    faulthandler.disable()

    try:
        _fileobj.close()
        os.unlink(_fileobj.name)
    except OSError:
        pass
    _fileobj = None


@atexit.register
def _at_exit():
    disable()


def raise_and_clear_error():
    """Raises an error if there is one. Calling this will clear the error
    so a second call won't do anything.

    enable() needs to be called first.

    Raises:
        FaultHandlerCrash
    """

    global _fileobj

    if _fileobj is None:
        return

    try:
        _fileobj.seek(0)
        text = _fileobj.read().decode("utf-8", "replace").strip()
        _fileobj.seek(0)
        _fileobj.truncate()
    except OSError:
        print_exc()
    else:
        if text:
            raise FaultHandlerCrash(text)


def crash():
    """Makes the process segfault. For testing purposes"""

    if os.name == "nt":
        i = ctypes.c_char(b"a")
        j = ctypes.pointer(i)
        c = 0
        while True:
            j[c] = b"a"
            c += 1
    else:
        ctypes.string_at(0)



# Copyright 2012,2013 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# TODO:
# RenameCommand
# FillTracknumberCommand

import ast
import os
import re
import shutil
import subprocess
import tempfile

from senf import fsn2text, text2fsn

from quodlibet import _
from quodlibet import util
from quodlibet.formats import EmbeddedImage, AudioFileError
from quodlibet.util.path import mtime
from quodlibet.pattern import Pattern, Error as PatternError
from quodlibet.util.tags import USER_TAGS, sortkey, MACHINE_TAGS
from quodlibet.util.tagsfrompath import TagsFromPattern

from .base import Command, CommandError
from .util import print_terse_table, copy_mtime, list_tags, print_table, get_editor_args


@Command.register
class ListCommand(Command):
    NAME = "list"
    DESCRIPTION = _("List tags")
    USAGE = "[-a] [-t] [-c <c1>,<c2>...] <file>"

    def _add_options(self, p):
        p.add_option("-t", "--terse", action="store_true", help=_("Print terse output"))
        p.add_option(
            "-c",
            "--columns",
            action="store",
            type="string",
            help=_("Columns to display and order in terse mode (%s)")
            % "desc,value,tag",
        )
        p.add_option(
            "-a", "--all", action="store_true", help=_("Also list programmatic tags")
        )

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError(_("Not enough arguments"))
        if len(args) > 1:
            raise CommandError(_("Too many arguments"))

        path = args[0]
        headers = [_("Description"), _("Value"), _("Tag")]
        nicks = ["desc", "value", "tag"]

        if not options.columns:
            order = nicks
        else:
            order = [n.strip() for n in options.columns.split(",")]

        song = self.load_song(path)
        tags = list_tags(song, machine=options.all, terse=options.terse)

        if options.terse:
            print_terse_table(tags, nicks, order)
        else:
            print_table(tags, headers, nicks, order)


@Command.register
class TagsCommand(Command):
    NAME = "tags"
    DESCRIPTION = _("List all common tags")
    USAGE = "[-t] [-c <c1>,<c2>...]"

    def _add_options(self, p):
        p.add_option("-t", "--terse", action="store_true", help=_("Print terse output"))
        p.add_option(
            "-c",
            "--columns",
            action="store",
            type="string",
            help=_("Columns to display and order in terse mode (%s)") % "tag,desc",
        )
        p.add_option(
            "-a", "--all", action="store_true", help=_("Also list programmatic tags")
        )

    def _execute(self, options, args):
        if len(args) != 0:
            raise CommandError(_("Too many arguments"))

        headers = [_("Tag"), _("Description")]
        nicks = ["tag", "desc"]

        if not options.columns:
            order = nicks
        else:
            order = [n.strip() for n in options.columns.split(",")]

        tag_names = list(USER_TAGS)
        if options.all:
            tag_names.extend(MACHINE_TAGS)

        tags = []
        for key in tag_names:
            tags.append((key, util.tag(key)))
        tags.sort()

        if not options.terse:
            print_table(tags, headers, nicks, order)
        else:
            print_terse_table(tags, nicks, order)


@Command.register
class CopyCommand(Command):
    NAME = "copy"
    DESCRIPTION = _("Copy tags from one file to another")
    USAGE = "[--dry-run] [--ignore-errors] <source> <dest>"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help=_("Show changes, don't apply them")
        )
        p.add_option(
            "--ignore-errors",
            action="store_true",
            help=_("Skip tags that can't be written"),
        )

    def _execute(self, options, args):
        if len(args) < 2:
            raise CommandError(_("Not enough arguments"))
        if len(args) > 2:
            raise CommandError(_("Too many arguments"))

        if options.dry_run:
            self.verbose = True

        source_path = args[0]
        dest_path = args[1]

        source = self.load_song(source_path)
        dest = self.load_song(dest_path)

        for key in source.realkeys():
            self.log(f"Copy {key!r}")
            if not options.ignore_errors and not dest.can_change(key):
                raise CommandError(
                    _("Can't copy tag {tagname} to file: {filename}").format(
                        tagname=repr(key), filename=repr(dest_path)
                    )
                )
            for value in source.list(key):
                dest.add(key, value)

        if not options.dry_run:
            self.save_songs([dest])


@Command.register
class EditCommand(Command):
    NAME = "edit"
    DESCRIPTION = _("Edit tags in a text editor")
    USAGE = "[--dry-run] <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help=_("Show changes, don't apply them")
        )

    def _song_to_text(self, song):
        # to text
        lines = [
            "File: {!r}".format(fsn2text(song("~filename"))),
            "",
        ]
        for key in sorted(song.realkeys(), key=sortkey):
            for value in song.list(key):
                lines.append(f"  {key}={value}")

        return "\n".join(lines + [""])

    def _songs_to_text(self, songs):
        header = [
            "# Lines before the first 'File:' statement, or"
            " lines that are empty or start with '#' will be ignored.",
            "",
        ]
        return "\n".join(header + [self._song_to_text(song) for song in songs])

    def _text_to_song(self, text, song):
        assert isinstance(text, str)

        # parse
        tags = {}
        for line in text.splitlines():
            if not line.strip() or line.startswith("#"):
                continue
            try:
                key, value = line.strip().split("=", 1)
            except ValueError:
                continue

            tags.setdefault(key, []).append(value)

        # apply changes, sort to always have the same output
        for key in sorted(song.realkeys(), key=sortkey):
            new = tags.pop(key, [])
            old = song.list(key)
            for value in old:
                if value not in new:
                    self.log(f"Remove {key}={value}")
                    song.remove(key, value)
            for value in new:
                if value not in old:
                    self.log(f"Add {key}={value}")
                    song.add(key, value)

        for key, values in tags.items():
            if not song.can_change(key):
                raise CommandError(
                    "Can't change key '%(key-name)s'." % {"key-name": key}
                )
            for value in values:
                self.log(f"Add {key}={value}")
                song.add(key, value)

    def _text_to_songs(self, text, songs):
        # remove comments
        text = re.sub(r"^#.*", "", text, count=0, flags=re.MULTILINE)
        # remove empty lines
        text = re.sub(r"(\r?\n){2,}", "\n", text.strip())
        _, *texts = re.split(r"^File:\s+", text, maxsplit=0, flags=re.MULTILINE)

        for text in texts:
            filename, *lines = text.splitlines()
            filename = text2fsn(ast.literal_eval(filename))
            text = "\n".join(lines)

            song = next((song for song in songs if song("~filename") == filename), None)
            if not song:
                raise CommandError(f"No match for {filename!r}.")

            self.log(f"Update song: {filename!r}")
            self._text_to_song(text, song)

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError(_("Not enough arguments"))

        songs = [self.load_song(path) for path in args]
        dump = self._songs_to_text(songs).encode("utf-8")

        # write to tmp file
        fd, path = tempfile.mkstemp(suffix=".txt")

        try:
            try:
                os.write(fd, dump)
            finally:
                os.close(fd)

            # XXX: copy mtime here so we can test for changes in tests by
            # setting a out of date mtime on the source song file
            copy_mtime(args[0], path)

            # only parse the result if the editor returns 0 and the mtime has
            # changed
            old_mtime = mtime(path)

            editor_args = get_editor_args()
            self.log(f"Using editor: {editor_args!r}")

            try:
                subprocess.check_call(editor_args + [path])
            except subprocess.CalledProcessError as e:
                self.log(str(e))
                raise CommandError(_("Editing aborted")) from e
            except OSError as e:
                self.log(str(e))
                raise CommandError(
                    _("Starting text editor '%(editor-name)s' failed.")
                    % {"editor-name": editor_args[0]}
                ) from e

            was_changed = mtime(path) != old_mtime
            if not was_changed:
                raise CommandError(_("No changes detected"))

            with open(path, "rb") as h:
                data = h.read()

        finally:
            os.unlink(path)

        try:
            text = data.decode("utf-8")
        except ValueError as e:
            raise CommandError(f"Invalid data: {e!r}") from e

        if options.dry_run:
            self.verbose = True
        self._text_to_songs(text, songs)

        if not options.dry_run:
            self.save_songs(songs)


@Command.register
class SetCommand(Command):
    NAME = "set"
    DESCRIPTION = _("Set a tag and remove existing values")
    USAGE = "[--dry-run] <tag> <value> <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help=_("Show changes, don't apply them")
        )

    def _execute(self, options, args):
        if len(args) < 3:
            raise CommandError(_("Not enough arguments"))

        tag = fsn2text(args[0])
        value = fsn2text(args[1])
        paths = args[2:]

        songs = []
        for path in paths:
            song = self.load_song(path)

            if not song.can_change(tag):
                vars = {
                    "tag": tag,
                    "format": type(song).format,
                    "file": song("~filename"),
                }
                raise CommandError(
                    _("Can not set %(tag)r for %(format)s file %(file)r") % vars
                )

            self.log(f"Set {value!r} to {tag!r}")
            if tag in song:
                del song[tag]
            song.add(tag, value)
            songs.append(song)

        if not options.dry_run:
            self.save_songs(songs)


@Command.register
class ClearCommand(Command):
    NAME = "clear"
    DESCRIPTION = _("Remove tags")
    USAGE = "[--dry-run] [-a | -e <pattern> | <tag>] <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help=_("Show changes, don't apply them")
        )
        p.add_option(
            "-e",
            "--regexp",
            action="store",
            type="string",
            help=_("Value is a regular expression"),
        )
        p.add_option("-a", "--all", action="store_true", help=_("Remove all tags"))

    def _execute(self, options, args):
        if options.all and options.regexp is not None:
            raise CommandError(_("Can't combine '--all' with '--regexp'"))

        if options.regexp is not None or options.all:
            if len(args) < 1:
                raise CommandError(_("Not enough arguments"))
            paths = args
        else:
            if len(args) < 2:
                raise CommandError(_("Not enough arguments"))
            paths = args[1:]

        if options.dry_run:
            self.verbose = True

        songs = []
        for path in paths:
            song = self.load_song(path)

            tags = []
            realkeys = song.realkeys()
            if options.all:
                tags.extend(realkeys)
            elif options.regexp is not None:
                e = re.compile(options.regexp)
                tags.extend(filter(e.match, realkeys))
            else:
                tag = args[0]
                if tag in realkeys:
                    tags.append(tag)

            for tag in tags:
                self.log(f"Remove tag {tag!r}")
                if not song.can_change(tag):
                    raise CommandError(
                        _("Can't remove {tagname} from {filename}").format(
                            tagname=repr(tag), filename=repr(path)
                        )
                    )
                del song[tag]

            if tags:
                songs.append(song)

        if not options.dry_run:
            self.save_songs(songs)


@Command.register
class RemoveCommand(Command):
    NAME = "remove"
    DESCRIPTION = _("Remove a tag value")
    USAGE = "[--dry-run] <tag> [-e <pattern> | <value>] <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help=_("Show changes, don't apply them")
        )
        p.add_option(
            "-e",
            "--regexp",
            action="store",
            type="string",
            help=_("Value is a regular expression"),
        )

    def _execute(self, options, args):
        if options.regexp is None:
            if len(args) < 3:
                raise CommandError(_("Not enough arguments"))
        else:
            if len(args) < 2:
                raise CommandError(_("Not enough arguments"))

        if options.dry_run:
            self.verbose = True

        tag = args[0]
        if options.regexp is not None:
            match = re.compile(options.regexp).match
            paths = args[1:]
        else:
            value = args[1]
            paths = args[2:]

            def match(v):
                return v == value

        songs = []
        for path in paths:
            song = self.load_song(path)

            if tag not in song:
                continue

            for v in song.list(tag):
                if match(v):
                    self.log(f"Remove {v!r} from {tag!r}")
                    song.remove(tag, v)
            songs.append(song)

        if not options.dry_run:
            self.save_songs(songs)


@Command.register
class AddCommand(Command):
    NAME = "add"
    DESCRIPTION = _("Add a tag value")
    USAGE = "<tag> <value> <file> [<files>]"

    def _execute(self, options, args):
        if len(args) < 3:
            raise CommandError(_("Not enough arguments"))

        tag = fsn2text(args[0])
        value = fsn2text(args[1])
        paths = args[2:]

        songs = []
        for path in paths:
            song = self.load_song(path)

            if not song.can_change(tag):
                raise CommandError(_("Can not set %r") % tag)

            self.log(f"Add {value!r} to {tag!r}")
            song.add(tag, value)
            songs.append(song)

        self.save_songs(songs)


@Command.register
class InfoCommand(Command):
    NAME = "info"
    DESCRIPTION = _("List file information")
    USAGE = "[-t] [-c <c1>,<c2>...] <file>"

    def _add_options(self, p):
        p.add_option("-t", "--terse", action="store_true", help=_("Print terse output"))
        p.add_option(
            "-c",
            "--columns",
            action="store",
            type="string",
            help=_("Columns to display and order in terse mode (%s)") % "desc,value",
        )

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError(_("Not enough arguments"))
        if len(args) > 1:
            raise CommandError(_("Too many arguments"))

        path = args[0]
        song = self.load_song(path)

        headers = [_("Description"), _("Value")]
        nicks = ["desc", "value"]

        if not options.columns:
            order = nicks
        else:
            order = [n.strip() for n in options.columns.split(",")]

        if not options.terse:
            tags = []
            for key in [
                "~format",
                "~codec",
                "~encoding",
                "~length",
                "~bitrate",
                "~filesize",
            ]:
                tags.append((util.tag(key), str(song.comma(key))))

            print_table(tags, headers, nicks, order)
        else:
            tags = []
            for key in [
                "~format",
                "~codec",
                "~encoding",
                "~#length",
                "~#bitrate",
                "~#filesize",
            ]:
                tags.append((key.lstrip("#~"), str(song(key))))

            print_terse_table(tags, nicks, order)


@Command.register
class ImageSetCommand(Command):
    NAME = "image-set"
    DESCRIPTION = _(
        "Set the provided image as primary embedded image and "
        "remove all other embedded images"
    )
    USAGE = "<image-file> <file> [<files>]"

    def _execute(self, options, args):
        if len(args) < 2:
            raise CommandError(_("Not enough arguments"))

        image_path = args[0]
        paths = args[1:]

        image = EmbeddedImage.from_path(image_path)
        if not image:
            raise CommandError(_("Failed to load image file: %r") % image_path)

        songs = [self.load_song(p) for p in paths]

        for song in songs:
            if not song.can_change_images:
                raise CommandError(
                    _(
                        "Image editing not supported for %(file_name)s "
                        "(%(file_format)s)"
                    )
                    % {"file_name": song("~filename"), "file_format": song("~format")}
                )

        for song in songs:
            try:
                song.set_image(image)
            except AudioFileError as e:
                raise CommandError(e) from e


@Command.register
class ImageClearCommand(Command):
    NAME = "image-clear"
    DESCRIPTION = _("Remove all embedded images")
    USAGE = "<file> [<files>]"

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError(_("Not enough arguments"))

        paths = args
        songs = [self.load_song(p) for p in paths]

        for song in songs:
            if not song.can_change_images:
                raise CommandError(
                    _(
                        "Image editing not supported for %(file_name)s "
                        "(%(file_format)s)"
                    )
                    % {"file_name": song("~filename"), "file_format": song("~format")}
                )

        for song in songs:
            try:
                song.clear_images()
            except AudioFileError as e:
                raise CommandError(e) from e


@Command.register
class ImageExtractCommand(Command):
    NAME = "image-extract"
    DESCRIPTION = _("Extract embedded images to %(filepath)s") % {
        "filepath": "<destination>/<filename>-<index>.(jpeg|png|..)"
    }
    USAGE = "[--dry-run] [--primary] [-d <destination>] <file> [<files>]"

    def _add_options(self, p):
        p.add_option("--dry-run", action="store_true", help="don't save images")
        p.add_option(
            "--primary", action="store_true", help="only extract the primary image"
        )
        p.add_option(
            "-d",
            "--destination",
            action="store",
            type="string",
            help=_(
                "Path to where the images will be saved to "
                "(defaults to the working directory)"
            ),
        )

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError(_("Not enough arguments"))

        # dry run implies verbose
        if options.dry_run:
            self.verbose = True

        paths = args
        for path in paths:
            song = self.load_song(path)

            # get the primary one or all of them
            if options.primary:
                image = song.get_primary_image()
                images = [image] if image else []
            else:
                images = song.get_images()

            self.log(f"Images for {path!r}: {images!r}")

            if not images:
                continue

            # get the basename from the song without the extension
            basename = os.path.basename(path)
            name = os.path.splitext(basename)[0]

            # at least two places, but same length for all images
            number_pattern = "%%0%dd" % (max(2, len(images) - 1))

            for i, image in enumerate(images):
                # get a appropriate file extension or use fallback
                extensions = image.extensions
                ext = extensions[0] if extensions else ".image"

                if options.primary:
                    # mysong.mp3 -> mysong.jpeg
                    filename = f"{name}.{ext}"
                else:
                    # mysong.mp3 -> mysong-00.jpeg
                    pattern = "%s-" + number_pattern + ".%s"
                    filename = pattern % (name, i, ext)

                if options.destination is not None:
                    filename = os.path.join(options.destination, filename)

                self.log(f"Saving image {filename!r}")
                if not options.dry_run:
                    with open(filename, "wb") as h:
                        shutil.copyfileobj(image.file, h)


# @Command.register
class RenameCommand(Command):
    NAME = "rename"
    DESCRIPTION = _("Rename files based on tags")
    USAGE = "[--dry-run] <pattern> <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help="show changes, don't apply them"
        )

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError("Not enough arguments")


@Command.register
class FillCommand(Command):
    NAME = "fill"
    DESCRIPTION = _("Fill tags based on the file path")
    USAGE = "[--dry-run] <pattern> <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help="show changes, don't apply them"
        )

    def _execute(self, options, args):
        if len(args) < 2:
            raise CommandError("Not enough arguments")

        pattern_text = args[0]
        self.log(f"Using pattern: {pattern_text!r}")
        paths = args[1:]

        pattern = TagsFromPattern(pattern_text)

        songs = []
        for path in paths:
            song = self.load_song(path)
            for header in pattern.headers:
                if not song.can_change(header):
                    raise CommandError(_("Can not set %r") % header)
            songs.append(song)

        if options.dry_run:
            self.__preview(pattern, songs)
        else:
            self.__apply(pattern, songs)

    def __apply(self, pattern, songs):
        for song in songs:
            match = pattern.match(song)
            self.log("{!r}: {!r}".format(song("~basename"), match))
            for header in pattern.headers:
                if header in match:
                    value = match[header]
                    song[header] = value

        self.save_songs(songs)

    def __preview(self, pattern, songs):
        rows = []
        for song in songs:
            match = pattern.match(song)
            row = [fsn2text(song("~basename"))]
            for header in pattern.headers:
                row.append(match.get(header, ""))
            rows.append(row)

        headers = [_("File")] + pattern.headers
        nicks = ["file"] + pattern.headers
        print_table(rows, headers, nicks, nicks)


# @Command.register
class FillTracknumberCommand(Command):
    NAME = "fill-tracknumber"
    DESCRIPTION = _("Fill tracknumbers for all files")
    USAGE = "[--dry-run] [--start] [--total] <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "--dry-run", action="store_true", help="show changes, don't apply them"
        )
        p.add_option("--start", action="store_true", help="tracknumber to start with")
        p.add_option("--total", action="store_true", help="total number of tracks")

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError("Not enough arguments")


@Command.register
class PrintCommand(Command):
    NAME = "print"
    DESCRIPTION = _("Print tags based on the given pattern")
    USAGE = "[-p <pattern>] <file> [<files>]"

    def _add_options(self, p):
        p.add_option(
            "-p",
            "--pattern",
            action="store",
            type="string",
            help="use a custom pattern",
        )

    def _execute(self, options, args):
        if len(args) < 1:
            raise CommandError("Not enough arguments")

        pattern = options.pattern
        if pattern is None:
            pattern = "<artist~album~tracknumber~title>"

        self.log(f"Using pattern: {pattern!r}")

        try:
            pattern = Pattern(pattern)
        except PatternError as e:
            raise CommandError(f"Invalid pattern: {pattern!r}") from e

        paths = args
        error = False
        for path in paths:
            try:
                util.print_(pattern % self.load_song(path))
            except CommandError:
                error = True

        if error:
            raise CommandError("One or more files failed to load.")


@Command.register
class HelpCommand(Command):
    NAME = "help"
    DESCRIPTION = _("Display help information")
    USAGE = "[<command>]"

    def _execute(self, options, args):
        if len(args) > 1:
            raise CommandError("Too many arguments")

        for cmd in Command.COMMANDS:
            if cmd.NAME == args[0]:
                cmd(self._main_cmd).print_help()
                break
        else:
            raise CommandError("Unknown command")



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
from collections.abc import Iterable, Generator

import quodlibet
from quodlibet import print_d, print_w, print_e, ngettext, _
from quodlibet.formats import AudioFile
from quodlibet.library.base import Library
from quodlibet.util.collection import Playlist, XSPFBackedPlaylist, FileBackedPlaylist
from quodlibet.util.path import is_hidden
from senf import text2fsn, _fsnative, fsnative

_DEFAULT_PLAYLIST_DIR = text2fsn(os.path.join(quodlibet.get_user_dir(), "playlists"))
"""Directory for playlist files"""


_MIN_NON_EMPTY_PL_BYTES = 4
"""Arbitrary minimum file size for a legacy non-empty playlist file"""


class PlaylistLibrary(Library[str, Playlist]):
    """A PlaylistLibrary listens to a SongLibrary, and keeps tracks of playlists
    of these songs.

    The library behaves like a dictionary: the keys are playlist names,
    the values are Playlist objects.
    """

    def __init__(self, library: Library, pl_dir: _fsnative = _DEFAULT_PLAYLIST_DIR):
        self.librarian = None
        super().__init__(f"{type(self).__name__} for {library._name}")
        print_d(f"Initializing Playlist Library {self} to watch {library._name!r}")
        self.pl_dir = pl_dir
        if library is None:
            raise ValueError("Need a library to listen to")
        self._library = library
        self._read_playlists(library)

        self._rsig = library.connect("removed", self.__songs_removed)
        self._csig = library.connect("changed", self.__songs_changed)

    def _read_playlists(self, library) -> None:
        print_d(f"Reading playlist directory {self.pl_dir} (library: {library})")
        try:
            fns = os.listdir(self.pl_dir)
        except FileNotFoundError as e:
            print_w(f"No playlist dir found in {self.pl_dir!r}, creating. ({e})")
            os.mkdir(self.pl_dir)
            fns = []

        # Populate this library by relying on existing signal passing.
        # Weird, but allows keeping the logic in one place
        failed = []
        for fn in fns:
            full_path = os.path.join(self.pl_dir, fn)
            if os.path.isdir(full_path):
                continue
            if is_hidden(fsnative(fn)):
                print_d(f"Ignoring hidden file {fn!r}")
                continue
            try:
                XSPFBackedPlaylist(self.pl_dir, fn, songs_lib=library, pl_lib=self)
            except TypeError as e:
                # Don't add to library - it's temporary
                legacy = FileBackedPlaylist(
                    self.pl_dir, fn, songs_lib=library, pl_lib=None
                )
                if not len(legacy):
                    try:
                        size = os.stat(legacy._last_fn).st_size
                        if size >= _MIN_NON_EMPTY_PL_BYTES:
                            data = {"filename": fn, "size": size / 1024}
                            print_w(
                                _(
                                    "No library songs found in legacy playlist "
                                    "%(filename)r (of size %(size).1f kB)."
                                )
                                % data
                                + " "
                                + _(
                                    "Have you changed library root dir(s), "
                                    "but not this playlist?"
                                )
                            )
                            continue
                    except OSError:
                        print_e(f"Problem reading {legacy._last_fn!r}")
                        continue
                    finally:
                        failed.append(fn)
                print_w(f"Converting {fn!r} to XSPF format ({e})")
                XSPFBackedPlaylist.from_playlist(legacy, songs_lib=library, pl_lib=self)
            except OSError:
                print_w(f"Invalid Playlist {fn!r}")
                failed.append(fn)
        if failed:
            total = len(failed)
            print_e(
                ngettext(
                    "%d playlist failed to convert",
                    "%d playlists failed to convert",
                    total,
                )
                % len(failed)
            )

    def create(self, name_base: str | None = None) -> Playlist:
        if name_base:
            return XSPFBackedPlaylist.new(
                self.pl_dir, name_base, songs_lib=self._library, pl_lib=self
            )
        return XSPFBackedPlaylist.new(self.pl_dir, songs_lib=self._library, pl_lib=self)

    def create_from_songs(self, songs: Iterable[AudioFile], title=None) -> Playlist:
        """Creates a playlist visible to this library"""
        return XSPFBackedPlaylist.from_songs(
            self.pl_dir, songs, title=title, songs_lib=self._library, pl_lib=self
        )

    def destroy(self):
        for sig in [self._rsig, self._csig]:
            self._library.disconnect(sig)

    def playlists_featuring(self, song: AudioFile) -> Generator[Playlist, None, None]:
        """Returns a generator yielding playlists in which this song appears"""
        return (pl for pl in self if song in pl._list)

    def __songs_removed(self, library, songs):
        print_d(
            f"Removing {len(songs)} song(s) "
            f"across {len(self)} playlist(s) in {self}"
        )
        changed = {pl for pl in self if pl.remove_songs(songs)}
        if changed:
            for pl in changed:
                pl.write()
            self.changed(changed)

    def rename(self, playlist: Playlist, old_key: str):
        """Handle the renaming (and thus re-keying) of contained items"""
        if playlist not in self and old_key in self.keys():
            del self._contents[old_key]
            self._contents[playlist.key] = playlist
            self.emit("changed", [playlist])
        else:
            print_w(f"Can't rename {old_key} -> {playlist.key}")

    def __songs_changed(self, library, songs) -> None:
        # Q: what if the changes are entirely due to changes *from* this library?
        # A: seems safest to still emit 'changed' as collections can cache metadata etc
        changed = set()
        for playlist in self:
            for song in songs:
                if song in playlist.songs:
                    changed.add(playlist)
                    # It's definitely changed now, nothing else is interesting
                    break
        if changed:
            # TODO: only write if anything *persisted* changes (#3622)
            #  i.e. not internal stuff (notably: ~playlists itself)
            for pl in changed:
                pl.finalize()
                pl.write()
            self.changed(changed)

    def recreate(self, playlist: Playlist, songs: Iterable[AudioFile]):
        """Keep a playlist but entirely replace its contents
        This is useful for applying new external sorting etc"""
        playlist._list.clear()
        playlist._list.extend(songs)
        playlist.finalize()
        playlist.write()
        self.changed([playlist])



# Copyright 2013 Christoph Reiter
#      2013-2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _, print_d, ngettext
from quodlibet import app
from quodlibet import util
from quodlibet.plugins.songshelpers import any_song, has_writable_image
from quodlibet.qltk.chooser import _get_chooser
from quodlibet.qltk.x import MenuItem
from quodlibet.qltk import Icons
from quodlibet.qltk.wlw import WritingWindow
from quodlibet.qltk._editutils import WriteFailedError
from quodlibet.formats import EmbeddedImage, AudioFileError, AudioFile
from quodlibet.plugins.songsmenu import SongsMenuPlugin


class EditEmbedded(SongsMenuPlugin):
    PLUGIN_ID = "embedded_edit"
    PLUGIN_NAME = _("Edit Embedded Images")
    PLUGIN_DESC = _("Adds, removes or replaces embedded images.")
    PLUGIN_ICON = Icons.INSERT_IMAGE

    plugin_handles = any_song(has_writable_image)
    """if any song supports editing, we are active"""

    def __init__(self, songs, *args, **kwargs):
        super().__init__(songs, *args, **kwargs)
        self.__menu = Gtk.Menu()
        self._init_submenu_items(self.__menu, songs)
        self.set_submenu(self.__menu)

    def __remove_images(self, menu_item, songs):
        win = WritingWindow(self.plugin_window, len(songs))
        win.show()

        for song in songs:
            if song.has_images and song.can_change_images:
                try:
                    song.clear_images()
                except AudioFileError:
                    util.print_exc()
                    WriteFailedError(win, song).run()

            if win.step():
                break

        win.destroy()
        self.plugin_finish()

    def __set_image(self, menu_item, songs):
        win = WritingWindow(self.plugin_window, len(songs))
        win.show()

        for song in songs:
            if song.can_change_images:
                fileobj = app.cover_manager.get_cover(song)
                if fileobj:
                    path = fileobj.name
                    image = EmbeddedImage.from_path(path)
                    if image:
                        _set_image_or_throw(image, song, win)
            if win.step():
                break

        win.destroy()
        self.plugin_finish()

    def __choose_image(self, menu_item, songs):
        dialog = _get_chooser(_("_Embed"), _("_Cancel"))
        total = len(songs)
        msg = (
            ngettext(
                "Choose image to embed in the %d track",
                "Choose image to embed in %d tracks",
                total,
            )
            % total
        )
        dialog.set_title(msg)
        response = dialog.run()
        path = dialog.get_filename()
        dialog.destroy()
        if response != Gtk.ResponseType.ACCEPT:
            print_d("User cancelled image embedding")
            return

        win = WritingWindow(self.plugin_window, len(songs))
        win.show()

        for song in songs:
            if song.can_change_images:
                image = EmbeddedImage.from_path(path)
                if image:
                    _set_image_or_throw(image, song, win)

            if win.step():
                break

        win.destroy()
        self.plugin_finish()

    def _init_submenu_items(self, menu, songs):
        remove_item = MenuItem(_("_Remove all Images"), Icons.EDIT_DELETE)
        remove_item.connect("activate", self.__remove_images, songs)
        remove_item.set_sensitive(any(song.has_images for song in songs))
        menu.append(remove_item)

        set_item = MenuItem(_("_Embed Current Image"), Icons.EDIT_PASTE)
        set_item.connect("activate", self.__set_image, songs)
        menu.append(set_item)

        set_item = MenuItem(_("_Choose Image…"), Icons.DOCUMENT_OPEN)
        set_item.connect("activate", self.__choose_image, songs)
        menu.append(set_item)

        menu.show_all()
        return menu

    def plugin_songs(self, songs):
        return True


def _set_image_or_throw(image: EmbeddedImage, song: AudioFile, win: Gtk.Window) -> None:
    try:
        song.set_image(image)
    except AudioFileError:
        util.print_exc()
        WriteFailedError(win, song).run()



# Copyright 2018 Phidica Veia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sqlite3

from gi.repository import Gtk
from senf import uri2fsn, text2fsn

from quodlibet import _
from quodlibet import app
from quodlibet import ngettext
from quodlibet import util
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk import Icons
from quodlibet.qltk.msg import Message, WarningMessage, ErrorMessage
from quodlibet.util.path import normalize_path
from quodlibet.plugins.events import EventPlugin


class BansheeDBImporter:
    def __init__(self, library):
        self._library = library
        self._changed_songs = []

    def read(self, db):
        """Iterate through the database and import data for songs found in
        the library
        """

        # use the Row class for extracting rows
        db.row_factory = sqlite3.Row

        # iterate over all songs in the database
        # throws sqlite3.OperationalError if CoreTracks is not found
        for row in db.execute("SELECT * FROM CoreTracks"):
            try:
                filename = uri2fsn(row["Uri"])
            except ValueError:
                continue

            song = self._library.get(normalize_path(filename))
            if not song:
                continue

            has_changed = False

            # rating is stored as integer from 0 to 5
            b_rating = row["Rating"] / 5.0
            if b_rating != song("~#rating"):
                song["~#rating"] = b_rating
                has_changed = True

            # play count is stored as integer from 0
            if row["PlayCount"] != song("~#playcount"):
                # summing play counts would break on multiple imports
                song["~#playcount"] = row["PlayCount"]
                has_changed = True

            # skip count is stored as integer from 0
            if row["SkipCount"] != song("~#skipcount"):
                song["~#skipcount"] = row["SkipCount"]
                has_changed = True

            # timestamp is stored as integer or None
            if row["LastPlayedStamp"] is not None:
                value = row["LastPlayedStamp"]
                # keep timestamp if it is newer than what we had
                if value > song("~#lastplayed", 0):
                    song["~#lastplayed"] = value
                    has_changed = True

            if row["DateAddedStamp"] is not None:
                value = row["DateAddedStamp"]
                # keep timestamp if it is older than what we had
                if value < song("~#added", 0):
                    song["~#added"] = value
                    has_changed = True

            if has_changed:
                self._changed_songs.append(song)

    def finish(self):
        """Call at the end, also returns number of songs with data imported"""

        count = len(self._changed_songs)
        self._library.changed(self._changed_songs)
        self._changed_songs = []
        return count


def do_import(parent, library):
    db_path = os.path.expanduser(BansheeImport.USR_PATH)

    importer = BansheeDBImporter(library)
    try:
        db = sqlite3.connect(db_path)
        importer.read(db)
        db.close()
    except sqlite3.OperationalError:
        msg = _("Specified Banshee database is malformed or missing")
        WarningMessage(parent, BansheeImport.PLUGIN_NAME, msg).run()
    except Exception:
        util.print_exc()
        importer.finish()
        msg = _("Import Failed")
        # FIXME: don't depend on the plugin class here
        ErrorMessage(parent, BansheeImport.PLUGIN_NAME, msg).run()
    else:
        count = importer.finish()
        msg = (
            ngettext(
                "Successfully imported ratings and statistics for %d song",
                "Successfully imported ratings and statistics for %d songs",
                count,
            )
            % count
        )
        Message(Gtk.MessageType.INFO, parent, BansheeImport.PLUGIN_NAME, msg).run()


class BansheeImport(EventPlugin):
    PLUGIN_ID = "bansheeimport"
    PLUGIN_NAME = _("Banshee Import")
    PLUGIN_DESC = _("Imports ratings and song statistics from Banshee.")
    PLUGIN_ICON = Icons.DOCUMENT_OPEN

    DEF_PATH = "~/.config/banshee-1/banshee.db"
    USR_PATH = DEF_PATH

    def PluginPreferences(self, *args):
        grid = Gtk.Grid(row_spacing=6, column_spacing=6)

        label = Gtk.Label(label=_("_Database path:"), use_underline=True)
        label.set_alignment(0.0, 0.5)
        grid.attach(label, 0, 0, 1, 1)

        entry = UndoEntry()
        entry.set_hexpand(True)
        entry.set_text(BansheeImport.DEF_PATH)

        def path_activate(entry, *args):
            path = text2fsn(entry.get_text())
            if BansheeImport.USR_PATH != path:
                BansheeImport.USR_PATH = path

        entry.connect_after("activate", path_activate)
        entry.connect_after("focus-out-event", path_activate)
        grid.attach_next_to(entry, label, Gtk.PositionType.RIGHT, 1, 1)

        path_revert = Gtk.Button()
        path_revert.add(
            Gtk.Image.new_from_icon_name(Icons.DOCUMENT_REVERT, Gtk.IconSize.MENU)
        )

        def path_revert_cb(button, entry):
            entry.set_text(BansheeImport.DEF_PATH)
            entry.emit("activate")

        path_revert.connect("clicked", path_revert_cb, entry)
        grid.attach_next_to(path_revert, entry, Gtk.PositionType.RIGHT, 1, 1)

        button = Gtk.Button(label=_("Start Import"))

        def clicked_cb(button):
            do_import(button, app.library)

        button.connect("clicked", clicked_cb)

        box = Gtk.VBox(spacing=12)

        box.pack_start(grid, True, True, 0)
        box.pack_start(button, False, False, 0)

        return box



# Copyright 2016-20 Nick Boultbee
#              2017 Jason Heard
#              2017 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _, print_d
from quodlibet.order import OrderInOrder
from quodlibet.plugins import PluginConfig
from quodlibet.plugins.playorder import ShufflePlugin
from quodlibet.qltk import Icons

pconfig = PluginConfig("skip_songs")
pconfig.defaults.set("threshold", 0.0)


class SkipDisliked(ShufflePlugin, OrderInOrder):
    PLUGIN_ID = "skip_songs"
    PLUGIN_NAME = _("Skip Disliked Tracks")
    PLUGIN_ICON = Icons.GO_JUMP
    PLUGIN_DESC = _(
        "Adds a play order (shuffle) mode that plays in order, "
        "but skips tracks with a rating "
        "below (or equal to) a given threshold."
    )
    display_name = _("Skip disliked tracks")
    accelerated_name = _("Skip _disliked tracks")

    @classmethod
    def PluginPreferences(cls, window):
        vb = Gtk.VBox(spacing=10)
        vb.set_border_width(0)

        adj = Gtk.Adjustment.new(pconfig.getfloat("threshold"), 0, 1.0, 0.01, 0.01, 0.0)
        fb_spin = Gtk.SpinButton(adjustment=adj)
        fb_spin.set_digits(2)

        def on_changed(button):
            pconfig.set("threshold", button.get_value())

        fb_spin.connect("changed", on_changed)

        vb.add(fb_spin)
        vb.show_all()
        return vb

    def next(self, playlist, current):
        next_ = super().next(playlist, current)

        while next_ is not None and self._should_skip(playlist, next_):
            next_ = super().next(playlist, next_)

        return next_

    def previous(self, playlist, current):
        previous = super().previous(playlist, current)
        is_first = False

        while (
            not is_first
            and previous is not None
            and self._should_skip(playlist, previous)
        ):
            previous = super().previous(playlist, previous)
            is_first = playlist.get_path(previous).get_indices()[0] == 0

        return previous

    def _should_skip(self, playlist, song_iter):
        song = playlist.get_value(song_iter)

        if not song.has_rating:
            return False

        rating = song("~#rating")
        threshold = pconfig.getfloat("threshold")

        if rating <= threshold:
            print_d(f"Rating is {rating:f} <= {threshold:f}; skipping...")
            return True
        return False



# Copyright 2005 Joe Wreschnig, Michael Urman
#        2016-23 Nick Boultbee
#           2018 Peter Strulo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Pango

from quodlibet import config, print_d
from quodlibet import util
from quodlibet import browsers
from quodlibet import app
from quodlibet import _

from quodlibet.qltk.songlist import SongList
from quodlibet.qltk.x import ScrolledWindow, Action
from quodlibet.qltk import Icons
from quodlibet.qltk.window import Window, PersistentWindowMixin
from quodlibet.util.library import background_filter


class FilterMenu:
    MENU = """
    <menu action='Filters'>
        <menuitem action='FilterGenre' always-show-image='true'/>
        <menuitem action='FilterArtist' always-show-image='true'/>
        <menuitem action='FilterAlbum' always-show-image='true'/>
        <separator/>
        <menuitem action='RandomGenre' always-show-image='true'/>
        <menuitem action='RandomArtist' always-show-image='true'/>
        <menuitem action='RandomAlbum' always-show-image='true'/>
        <separator/>
        <menuitem action='All' always-show-image='true'/>
        <menuitem action='PlayedRecently' always-show-image='true'/>
        <menuitem action='AddedRecently' always-show-image='true'/>
        <menuitem action='TopRated' always-show-image='true'/>
    </menu>"""
    __OUTER_MENU = f"""
    <ui>
        <menubar name='Menu'>
            {MENU}
        </menubar>
    </ui>"""

    def __init__(self, library, player, ui=None):
        self._browser = None
        self._library = library
        self._player = player
        self._standalone = not ui

        ag = Gtk.ActionGroup.new("QuodLibetFilterActions")
        for name, icon_name, label, cb in [
            ("Filters", "", _("_Filters"), None),
            (
                "PlayedRecently",
                Icons.EDIT_FIND,
                _("Recently _Played"),
                self.__filter_menu_actions,
            ),
            (
                "AddedRecently",
                Icons.EDIT_FIND,
                _("Recently _Added"),
                self.__filter_menu_actions,
            ),
            ("TopRated", Icons.EDIT_FIND, _("_Top 40"), self.__filter_menu_actions),
            ("All", Icons.EDIT_FIND, _("All _Songs"), self.__filter_menu_actions),
        ]:
            action = Action(name=name, icon_name=icon_name, label=label)
            if cb:
                action.connect("activate", cb)
            ag.add_action(action)

        for tag_, lab in [
            ("genre", _("On Current _Genre(s)")),
            ("artist", _("On Current _Artist(s)")),
            ("album", _("On Current Al_bum")),
        ]:
            act = Action(
                name=f"Filter{util.capitalize(tag_)}",
                label=lab,
                icon_name=Icons.EDIT_SELECT_ALL,
            )
            act.connect("activate", self.__filter_on, tag_, None, player)
            ag.add_action(act)

        for tag_, accel, label in [
            ("genre", "G", _("Random _Genre")),
            ("artist", "T", _("Random _Artist")),
            ("album", "M", _("Random Al_bum")),
        ]:
            act = Action(
                name=f"Random{util.capitalize(tag_)}",
                label=label,
                icon_name=Icons.DIALOG_QUESTION,
            )
            act.connect("activate", self.__random, tag_)
            ag.add_action_with_accel(act, "<Primary>" + accel)

        if self._standalone:
            ui = Gtk.UIManager()
            ui.add_ui_from_string(self.__OUTER_MENU)
        ui.insert_action_group(ag, -1)
        self._ui = ui

        self._get_child_widget("TopRated").set_tooltip_text(
            _(
                "The 40 songs you've played most (more than 40 may "
                "be chosen if there are ties)"
            )
        )

        # https://git.gnome.org/browse/gtk+/commit/?id=b44df22895c79
        menu_item = self._get_child_widget("/Menu/Filters")
        if isinstance(menu_item, Gtk.ImageMenuItem):
            menu_item.set_image(None)

        self._player_id = player.connect("song-started", self._on_song_started)
        self.set_song(player.song)

        self._hide_menus()

    def destroy(self):
        if self._player:
            self._player.disconnect(self._player_id)
        self._player = None
        self._browser = None
        self._library = None

    def _on_song_started(self, player, song):
        self.set_song(song)

    def __random(self, item, key):
        self._browser.filter_random(key)

    def __filter_on(self, action, header, songs, player):
        # Fall back to the playing song
        if songs is None:
            if player.song:
                songs = [player.song]
            else:
                return

        self._browser.filter_on(songs, header)

    def __filter_menu_actions(self, menuitem):
        name = menuitem.get_name()

        if name == "PlayedRecently":
            self._make_query("#(lastplayed < 7 days ago)")
        elif name == "AddedRecently":
            self._make_query("#(added < 7 days ago)")
        elif name == "TopRated":
            bg = background_filter()
            songs = (bg and filter(bg, self._library)) or self._library
            songs = [song.get("~#playcount", 0) for song in songs]
            if len(songs) == 0:
                return
            songs.sort()
            if len(songs) < 40:
                self._make_query(f"#(playcount > {songs[0] - 1:d})")
            else:
                self._make_query(f"#(playcount > {songs[-40] - 1:d})")
        elif name == "All":
            self._browser.unfilter()

    def _make_query(self, query):
        assert isinstance(query, str)
        if self._browser.can_filter_text():
            self._browser.filter_text(query)
            self._browser.activate()

    def _hide_menus(self):
        menus = {
            "genre": [
                "FilterGenre",
                "RandomGenre",
            ],
            "artist": [
                "FilterArtist",
                "RandomArtist",
            ],
            "album": [
                "FilterAlbum",
                "RandomAlbum",
            ],
            None: [
                "PlayedRecently",
                "AddedRecently",
                "TopRated",
                "All",
            ],
        }

        for key, widget_names in menus.items():
            if self._browser:
                can_filter = self._browser.can_filter(key)
            else:
                can_filter = False
            for name in widget_names:
                self._get_child_widget(name).set_property("visible", can_filter)

    def set_browser(self, browser):
        self._browser = browser
        self._hide_menus()

    def set_song(self, song):
        for wid in ["FilterAlbum", "FilterArtist", "FilterGenre"]:
            self._get_child_widget(wid).set_sensitive(bool(song))

        if song:
            for h in ["genre", "artist", "album"]:
                widget = self._get_child_widget(f"Filter{h.capitalize()}")
                widget.set_sensitive(h in song)

    def _get_child_widget(self, name=None):
        path = "/Menu%s/Filters" % ("" if self._standalone else "/Browse")
        if name:
            path += "/" + name
        return self._ui.get_widget(path)

    def get_widget(self):
        path = "/Menu" if self._standalone else "/Menu/Browse"
        return self._ui.get_widget(path)

    def get_accel_group(self):
        return self._ui.get_accel_group()


class LibraryBrowser(Window, util.InstanceTracker, PersistentWindowMixin):
    @classmethod
    def open(cls, browser_cls, library, player):
        """Creates and shows a new browser instance"""

        browser = cls(browser_cls, library, player)
        browser.show()
        return browser

    @classmethod
    def save(cls):
        """See which browser windows are open and save their names
        so we can restore them on start.
        """

        config.set(
            "memory",
            "open_browsers",
            "\n".join(browser.name for browser in cls.instances()),
        )

    @classmethod
    def restore(cls, library, player):
        """restore saved browser windows"""

        value = config.get("memory", "open_browsers", "")
        for name in value.split():
            kind = browsers.get(name)
            browser = cls(kind, library, player)
            browser.show_maybe()

    def __init__(self, browser_cls, library, player):
        super().__init__(dialog=False)
        self._register_instance()
        self.name = browser_cls.__name__

        self.set_default_size(600, 400)
        self.enable_window_tracking("browser_" + self.name)
        self.set_title(browser_cls.name + " - Quod Libet")
        self.add(Gtk.VBox())

        view = SongList(library, update=True)
        view.info.connect("changed", self.__set_totals)
        self.songlist = view
        self.songlist.sortable = not browser_cls.can_reorder

        sw = ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.add(view)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        self.browser = browser = browser_cls(library)
        if browser.can_reorder:
            view.enable_drop()
        elif browser.dropped:
            view.enable_drop(False)
        if browser.accelerators:
            self.add_accel_group(browser.accelerators)

        self.__container = browser.pack(sw)
        self.get_child().pack_start(self.__container, True, True, 0)

        main = self.get_child()
        bottom = Gtk.HBox()
        main.pack_end(bottom, False, True, 0)

        self._filter_menu = filter_menu = FilterMenu(library, player)
        filter_menu.set_browser(self.browser)
        self.add_accel_group(filter_menu.get_accel_group())
        bottom.pack_start(filter_menu.get_widget(), False, True, 0)
        filter_menu.get_widget().show()

        self.__statusbar = Gtk.Label()
        self.__statusbar.set_alignment(1.0, 0.5)
        self.__statusbar.set_padding(6, 3)
        self.__statusbar.set_ellipsize(Pango.EllipsizeMode.START)
        bottom.pack_end(self.__statusbar, True, True, 0)
        self.__statusbar.show()
        bottom.show()

        browser.connect("songs-selected", self.__browser_cb)
        browser.finalize(False)
        view.connect("popup-menu", self._menu, library)
        view.connect("drag-data-received", self.__drag_data_recv)
        view.connect("row-activated", self.__enqueue, player)

        if browser.headers is not None:
            view.connect("columns-changed", self.__cols_changed, browser)
            self.__cols_changed(view, browser)
        sw.show_all()
        for c in self.get_child().get_children():
            c.show()
        self.get_child().show()

        self.connect("destroy", self._on_destroy)

    def _on_destroy(self, *args):
        self._filter_menu.destroy()

    def __browser_cb(self, browser, songs, sorted):
        if browser.background:
            bg = background_filter()
            if bg:
                songs = list(filter(bg, songs))
        print_d(f"Setting {len(songs)} songs...")
        self.songlist.set_songs(songs, sorted)

    def __enqueue(self, view, path, column, player):
        app.window.playlist.enqueue([view.get_model()[path][0]])
        if player.song is None:
            player.next()

    def __drag_data_recv(self, view, *args):
        if self.browser.can_reorder:
            songs = view.get_songs()
            self.browser.reordered(songs)
        view.clear_sort()

    def __cols_changed(self, view, browser):
        for header in view.get_columns():
            tag = header.header_name
            for t in util.tagsplit(tag):
                if t in browser.headers:
                    header.set_visible(True)
                    break
            else:
                header.set_visible(False)

    def _menu(self, view: SongList, library) -> bool:
        path, col = view.get_cursor()
        header = col.header_name
        menu = view.menu(header, self.browser, library)
        if menu is not None:
            view.popup_menu(menu, 0, Gtk.get_current_event_time())
        return True

    def __set_totals(self, info, songs):
        i = len(songs)
        length = sum(song.get("~#length", 0) for song in songs)
        t = self.browser.status_text(count=i, time=util.format_time_preferred(length))
        self.__statusbar.set_text(t)



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.aac import AAC

from ._audio import AudioFile, translate_errors

extensions = [".aac", ".adif", ".adts"]


class AACFile(AudioFile):
    """ADTS/ADIF files"""

    format = "AAC"
    mimes = ["audio/x-aac"]
    fill_length = True

    def __init__(self, filename):
        with translate_errors():
            audio = AAC(filename)
        self["~#length"] = audio.info.length
        self["~#bitrate"] = int(audio.info.bitrate / 1000)
        if audio.info.channels:
            self["~#channels"] = audio.info.channels
        self["~#samplerate"] = audio.info.sample_rate
        self.sanitize(filename)

    def write(self):
        pass

    def reload(self, *args):
        title = self.get("title")
        super().reload(*args)
        if title is not None:
            self.setdefault("title", title)

    def can_change(self, k=None):
        if k is None:
            return ["title"]
        return k == "title"


loader = AACFile
types = [AACFile]
